"""
    Billets - a module for OpenERP to help organise billets for events.
    Copyright (C) 2011  Andrew Miller <andrew@amxl.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import tools
from osv import fields,osv,orm
import wizard
import billet
import math
import random
import pooler

option_form = '''<?xml version="1.0"?>
<form string="Assignment options">
    <field name="event"/>
    <field name="override_automatic_prov"/>
    <field name="override_manual_prov"/>
    <field name="always_override_rejected"/>
    <field name="override_finalised"/>
    <field name="override_emailed"/>
    <field name="override_confirmed"/>
    <field name="always_override_cancelled"/>
</form>
'''
option_fields = { 
    'event': {'type':'many2one', 'string':'Event', 'required':True, 'relation':'billets.event'},
    'override_automatic_prov': {'type':'boolean','string':'Reassign automated provisional assignments?'},
    'override_manual_prov': {'type':'boolean','string':'Reassign manual provisional assignments?'},
    'always_override_rejected': {'type':'boolean','string':'Always reassign rejected assignments?'},
    'override_finalised': {'type':'boolean','string':'Reassign finalised assignments?'},
    'override_emailed': {'type':'boolean','string':'Reassign assignments marked as e-mailed?'},
    'override_confirmed': {'type':'boolean','string':'Reassign assignments marked as confirmed?'},
    'always_override_cancelled': {'type':'boolean','string':'Always reassign when the other party cancelled?'}
                }

result_form = '''<?xml version="1.0"?>
<form string="Assignment Result">
  <label string="Results:" colspan="4"/>
  <field name="resulttext" nolabel="1" colspan="4"/>
</form>
'''
result_fields = { 'resulttext' : { 'type':'text', 'readonly': True}}

def _do_defaults(self, cr, uid, data, context):
    data['always_override_rejected'] = True
    data['always_override_cancelled'] = True
    return data

class MatchObjectiveFunction:
    def __init__(self, optover):
        self.optover = optover
    def compare(self, a, b):
        e = 0.0
        for k in self.optover:
            if a[k] == 'flex' or (a[k] == 'flexno' and b[k] in ('defno', 'flexno', 'flex')) or (a[k] == 'flexyes' and b[k] in ('defyes', 'flexyes', 'flex')) or (a[k] == 'defno' and b[k] in ('defno', 'flexno', 'flex')) or (a[k] == 'defyes' and b[k] in ('defyes', 'flexyes', 'flex')):
                pass
            elif (a[k] == 'flexyes' and b[k] == 'flexno') or (a[k] == 'flexno' and b[k] == 'flexyes'):
                e += 1
            elif (a[k] in ('flexyes', 'flexno') or b[k] in ('flexyes', 'flexno')):
                e += 2
            else:
                e += 1000000
        return e
    def compareAll(self, a, b, pairs):
        tot = 0
        for i in range(0, len(a)):
            ai = a[i]
            if pairs[i] < len(b):
                bi = b[pairs[i]]
                tot += self.compare(ai, bi)
        return tot

def randomlyPermuteOne(pairs):
    pairs = list(pairs)
    lp = len(pairs)
    s1 = int(random.uniform(0, lp))
    s2 = int(random.uniform(0, lp))
    t = pairs[s1]
    pairs[s1] = pairs[s2]
    pairs[s2] = t
    return pairs

def doOptimisation(a, b, f):
    la = len(a)
    lb = len(b)
    pairs = range(0, max(la, lb))
    bestPairs = pairs
    lastEnergy = f.compareAll(a, b, pairs)
    bestEnergy = lastEnergy
    coolingSchedule = map(lambda x: x / 100000.0 * lastEnergy, range(100000, 0, -1))
    for temp in coolingSchedule:
        if bestEnergy == 0:
            return (bestPairs, bestEnergy)
        newPairs = randomlyPermuteOne(pairs)
        newEnergy = f.compareAll(a, b, newPairs)
        p = math.exp((lastEnergy - newEnergy) / temp)
        if (p > random.uniform(0, 1)):
            lastEnergy = newEnergy
            pairs = newPairs
            if lastEnergy < bestEnergy:
                bestEnergy = lastEnergy
                bestPairs = pairs
    return (bestPairs, bestEnergy)

def _do_assignment(self, cr, uid, data, context):
    # Build a list of all offers:
    ret = ''
    alwaysoverridesO = ''
    alwaysoverridesW = ''
    if data['always_override_rejected']:
        alwaysoverridesO += ' OR notifystatus=\'rejected\''
        alwaysoverridesW += ' OR notifystatus=\'rejected\''
    if data['always_override_cancelled']:
        alwaysoverridesO += ' OR EXISTS (SELECT * FROM billets_wanted AS w WHERE w.cancelled=true AND w.provider=billets_offered_slot.id)'
        alwaysoverridesW += ' OR EXISTS (SELECT * FROM billets_offered_slot AS o WHERE o.cancelled=true AND o.id=provider)'

    eid = data['form']['event']

    matchreqs = 'matchstatus=\'none\''
    if data['form']['override_automatic_prov']:
        matchreqs += ' OR matchstatus=\'provaut\''
    if data['form']['override_manual_prov']:
        matchreqs += ' OR matchstatus=\'provman\''
    if data['form']['override_finalised']:
        matchreqs += ' OR matchstatus=\'final\''
    notifyreqs = 'notifystatus=\'none\''
    if data['form']['override_emailed']:
        notifyreqs += ' OR notifystatus\'email\''
    if data['form']['override_confirmed']:
        notifyreqs += ' OR notifystatus\'confirm\''

    cr.execute('SELECT DISTINCT id,strict_vegan,low_carb,omnivore,lacto_ovo_veg,gluten_free,other_diets,female,male,transport FROM billets_offered_slot WHERE event=%s AND cancelled=false AND offer IS NOT NULL AND (((' + matchreqs + ') AND (' + notifyreqs + '))' + alwaysoverridesO + ')', [eid])
    slots = cr.dictfetchall()
    alldays = {}
    pool = pooler.get_pool(cr.dbname)
    for s in slots:
        for d in pool.get('billets.offered.slot').browse(cr, uid, [s['id']])[0].days:
            di = d.id
            s['day%d' % di] = 'flex'
            alldays[di] = 1
    for di in alldays.keys():
        for s in slots:
            s['day%d' % di] = 'defno'

    cr.execute('SELECT DISTINCT id,strict_vegan,low_carb,omnivore,lacto_ovo_veg,gluten_free,other_diets,gender,transport FROM billets_wanted WHERE event=%s AND cancelled=false AND (((' + matchreqs + ') AND (' + notifyreqs + '))' + alwaysoverridesW + ')', [eid])
    wanted = cr.dictfetchall()
    for w in wanted:
        if w['gender'] == 'male':
            w['male'] = 'defyes'
            w['female'] = 'defno'
        elif w['gender'] == 'female':
            w['female'] = 'defyes'
            w['male'] = 'defno'
        else:
            w['male'] = 'flex'
            w['female'] = 'flex'
        del w['gender']

    alldays = {}
    for w in wanted:
        for d in pool.get('billets.wanted').browse(cr, uid, [w['id']])[0].days:
            di = d.id
            w['day%d' % di] = 'defyes'
            alldays[di] = 1
    for di in alldays.keys():
        for w in wanted:
            w['day%d' % di] = 'flex'
            
    alldiff = {}
    for w in wanted:
        for k in w:
            if w[k] in ('defyes','defno', 'flexno', 'flexyes'):
                alldiff[k] = 1
    for s in slots:
        for k in s:
            if w[k] in ('defyes','defno', 'flexno', 'flexyes'):
                alldiff[k] = 1
    optover = alldiff.keys()
    objf = MatchObjectiveFunction(optover)
    (pairs, energy) = doOptimisation(wanted, slots, objf)
    ret += "Solution found with score %f (lower is better)\n" % energy
    idsToMatch = []
    unmatched = []
    againstRequirements = []
    againstPreferences = []

    for i in range(0, len(wanted)):
        wantedi = wanted[i]
        if pairs[i] >= len(slots):
            unmatched.append(wantedi['id'])
        else:
            sloti = slots[pairs[i]]
            s = objf.compare(wantedi, sloti)
            if s >= 1000000:
                againstRequirements.append(wantedi['id'])
            elif s > 0:
                againstPreferences.append(wantedi['id'])
            idsToMatch.append((wantedi['id'], sloti['id']))

    if len(unmatched) > 0:
        ret += "There were too few spaces, so the following people were not assigned a billet:\n"
        for w in pool.get('billets.wanted').browse(cr, uid, unmatched):
            ret += " * %s\n" % w.partner_id.name
    if len(againstRequirements) > 0:
        ret += "Even the best solution assigned the following people to accommodation against the stated requirements:\n"
        for w in pool.get('billets.wanted').browse(cr, uid, againstRequirements):
            ret += " * %s\n" % w.partner_id.name
    if len(againstPreferences) > 0:
        ret += "Even the best solution assigned the following people to accommodation against the stated preferences:\n"
        for wanted in pool.get('billets.wanted').browse(cr, uid, againstPreferences):
            ret += " * %s\n" % wanted.partner_id.name

    for wantedId, slotId in idsToMatch:
        cr.execute('UPDATE billets_wanted SET matchstatus=\'none\',notifystatus=\'none\',provider=NULL WHERE provider=%s',
                   [slotId])
        cr.execute('UPDATE billets_wanted SET matchstatus=\'provaut\',notifystatus=\'none\',provider=%s WHERE id=%s', [slotId, wantedId])
        cr.execute('UPDATE billets_offered_slot SET matchstatus=\'provaut\',notifystatus=\'none\' WHERE id=%s', [slotId])
    cr.commit()

    data['resulttext'] = ret
    return data

class wiz_assign_billets(wizard.interface):
    states = {
          'init': {
              'actions': [_do_defaults],
              'result': {'type':'form', 'arch':option_form, 'fields':option_fields, 'state':[('end','Cancel'),('draft_assignment', 'Start Assignment')]}
            },
          'draft_assignment': {
            'actions': [_do_assignment],
            'result': {'type':'form', 'arch':result_form, 'fields':result_fields, 'state':[('end','OK')]}
            }
        }
wiz_assign_billets('billets.assign')
